//*************************************************************************
//* -----------------
//* CDCHitMaker Class
//* -----------------
//* (Description)
//*	This set of classes makes hits.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCCluster:
//*	CDCClusterBuf:
//*	CDCClusterMaker:
//* (Provides)
//*	CDCHitMaker		: steering class
//*	CDCHitBuf		: hit data buffer class
//*	CDCHitLayer		: hit layer class
//*	CDCHitCell		: hit cell class
//*	CDCHitWire		: hit wire class
//*	CDCHit			: hit data class
//*	CDCTimeToXRelation	: x-t relation
//* (Update Record)
//*	1999/12/13  K.Fujii	Original version.
//*	2000/05/14  K.Fujii	Introduced layer/cell/wire structure.
//*	2000/09/30  K.Fujii	Very temporary treatment for large M.S.
//*				in Hiroshima experiment. Should be
//*				improved later.
//*	2000/10/02  K.Fujii	Fixed a bug in CDCHitBuf::MakeHits.
//*************************************************************************

#include "CDCHitMaker.h"
#include <fstream.h>

Double_t CDCXTDatum::fSigmaMS = 0.;

//_____________________________________________________________________
// ---------------------
// X-t Calibration Datum
// ---------------------
ClassImp(CDCXTDatum)

CDCXTDatum::CDCXTDatum(Int_t n, Double_t *x, Double_t *t) : fN(n), fNpar(0)
{
  	if (fN) {
  		fX = new Double_t [fN];
  		fT = new Double_t [fN];
		Int_t i;
		for (i = 0; i<fN; i++) {
			fX[i] = x[i];
			fT[i] = t[i];
		}
	} else {
		fX = 0;
		fT = 0;
	}
	fDXpar = 0; // DXpar has to be initialized by SetDXpar
}

CDCXTDatum::CDCXTDatum(CDCXTDatum &xt) : fN(xt.fN), fNpar(xt.fNpar)
{
	if (fN) {
  		fX = new Double_t [fN];
  		fT = new Double_t [fN];
		Int_t i;
		for (i = 0; i<fN; i++) {
			fX[i] = xt.fX[i];
			fT[i] = xt.fT[i];
		}
		fDXpar = new Double_t [fNpar];
		for (i = 0; i<4; i++) fDXpar[i] = xt.fDXpar[i];
	} else {
		fX = 0;
		fT = 0;
		fDXpar = 0;
	}
}

CDCXTDatum::~CDCXTDatum()
{
	if (fX) 	delete [] fX;
	if (fT) 	delete [] fT;
	if (fDXpar) 	delete [] fDXpar;
}

void CDCXTDatum::SetDXpar(Int_t n, Double_t *par)
{
	if (n <= 0) {
		cerr 	<< "////// ERROR in CDCXTDatum::SetDXpar" << endl
			<< "  Invalid No points = " << n << endl;
		return;
	}
	if (n > fNpar) {
		if (fDXpar) delete [] fDXpar;
		fNpar  = n;
		fDXpar = new Double_t [fNpar];
	}
	Int_t i;
	for (i = 0; i<fNpar; i++) fDXpar[i] = par[i];
}

void CDCXTDatum::SetSigmaMS(Double_t sigma)
{
	fSigmaMS = sigma;
}

Int_t CDCXTDatum::WhereIs(Double_t t)
{
	Int_t base = 0;
	Int_t last = fN - 2;
	if (last < 0) {
		cerr << "CDCTimeToXRelation::CalibDatum::WhereIs()"
		     << " Negative last = " << last << endl;
		return 0;
	}
	Int_t position;
	if (t <= fT[0] ) {
		position = 0;
	} else if (t >= fT[last]) {
		position = last;
	} else {
		while (last >= base) {
			position = (base+last)/2;
			if (fT[position] <= t && t < fT[position+1]) {
				break;
			} else if (fT[position+1] <= t) {
				base = position + 1;
			} else {
				last = position;
			}
		}
	}
	return position;
}

Double_t CDCXTDatum::X(Double_t t)
{
	Int_t n    = WhereIs(t);
	return fX[n] + ((fX[n+1]-fX[n])/(fT[n+1]-fT[n]))*(t-fT[n]);
}

Double_t CDCXTDatum::DX(Double_t x)
{
// START VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP
#if 0
	Double_t dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
				   + fDXpar[2]*x + fDXpar[3] );
#else
	Double_t dx = TMath::Sqrt( fDXpar[0]*TMath::Exp(-fDXpar[1]*x)
				   + fDXpar[2]*x + fDXpar[3]
				   + fSigmaMS*fSigmaMS);
#endif
// END VERY TEMPORARY TREATMENT FOR HIROSHIMA EXP
	return dx;
}


//_____________________________________________________________________
// --------------------
// X-t Calibration Data
// --------------------

ClassImp(CDCTimeToXRelation)

CDCTimeToXRelation::CDCTimeToXRelation(Int_t n, CDCXTDatum *xt, Int_t last)
{
	if (n) {
  		CalibDatum *datum = new CalibDatum(last);
  		datum->Add(xt);
  		Add(datum);
  		fCurr = datum;
  	}
}

CDCTimeToXRelation::CDCTimeToXRelation(Char_t * calibfile) 
{
	cout << " Going to open a calib file " << calibfile << endl;
	ifstream in;
	in.open(calibfile,ios::in);
	cout << " Opened " << calibfile << endl;
        Int_t last = 99999999;
        Int_t n;
        Double_t *x   = new Double_t [1000];
        Double_t *t   = new Double_t [1000];
        Double_t *par = new Double_t [100];
        Double_t mton = 1000.;
	while ((in >> last >> n)) {
		Int_t np, wireno;
		CalibDatum *calp = new CalibDatum(last);
		Int_t  iwir, i, idummy;
		Char_t cdummy[1000];
		Double_t ddummy;
		for (iwir = 0; iwir < n; iwir++) {
			in >> wireno >> np;
			for (i = 0; i < np; i++) {
				in >> idummy >> x[i] >> t[i];
				t[i] *= mton;
			}
			calp->Add(new CDCXTDatum(np, x, t));
		}
		for (iwir = 0; iwir < n; iwir++) {
			in >> wireno >> np;
			for (i = 0; i < np; i++) {
				in >> idummy >> cdummy >> par[i]
				   >> ddummy >> ddummy >> ddummy;
			}
			((CDCXTDatum *)calp->At(iwir))->SetDXpar(np, par);
		}
		Add(calp);
        }
	delete [] x;
	delete [] t;
	delete [] par;
	in.close();
}

CDCTimeToXRelation::~CDCTimeToXRelation() 
{
	TIter next(this);
	CalibDatum *calp;
	while ((calp = (CalibDatum *)next())) {
		calp->Delete();
	}
	Delete();
}

void CDCTimeToXRelation::Update()
{
  	Int_t run = gJSF->GetRunNumber();
  	TIter next(this);
  	CalibDatum *datum;
  	while ((datum = (CalibDatum *)next())) {
  		fCurr = datum;
  		if (run <= fCurr->GetLastRun()) break;
  	}	
}


//_____________________________________________________________________
// ---------
// Hit Datum
// ---------

ClassImp(CDCHit)

CDCHit::CDCHit(CDCCluster *cluster) :
  	fDone(kFALSE), fCluster(cluster),
  	fXdrift(0.), fDXdrift(0.),
	fWire(0) {}

CDCHit::CDCHit(CDCHit &h) : Lockable(h), 
	fDone(h.fDone), fCluster(h.fCluster), 
	fXdrift(h.fXdrift), fDXdrift(h.fDXdrift),
	fWire(h.fWire) {}

CDCHit::~CDCHit() {}

Int_t	CDCHit::GetLayerNo() const
{
	if (fWire) {
		return fWire->GetLayerNo();
	} else {
		Int_t layer, cell, wire;
		Double_t x, y;
		CDCHitWire::ChannelToWire(fCluster,layer,cell,wire,x,y);
		return layer;
	}
}

Int_t	CDCHit::GetCellNo () const
{
	if (fWire) {
		return fWire->GetCellNo();
	} else {
		Int_t layer, cell, wire;
		Double_t x, y;
		CDCHitWire::ChannelToWire(fCluster,layer,cell,wire,x,y);
		return cell;
	}
}

Int_t	CDCHit::GetWireNo () const
{
	if (fWire) {
		return fWire->GetWireNo();
	} else {
		Int_t layer, cell, wire;
		Double_t x, y;
		CDCHitWire::ChannelToWire(fCluster,layer,cell,wire,x,y);
		return wire;
	}
}
  
Double_t CDCHit::Xwire() const
{
	if (fWire) {
		return fWire->X();
	} else {
		Int_t layer, cell, wire;
		Double_t x, y;
		CDCHitWire::ChannelToWire(fCluster,layer,cell,wire,x,y);
		return x;
	}
}
  
Double_t CDCHit::Ywire() const
{
	if (fWire) {
		return fWire->Y();
	} else {
		Int_t layer, cell, wire;
		Double_t x, y;
		CDCHitWire::ChannelToWire(fCluster,layer,cell,wire,x,y);
		return y;
	}
}
  

Bool_t CDCHit::IsSortable() const { return kTRUE; }

Int_t CDCHit::Compare(TObject *obj)
{
	if (!fDone) Update();
	Double_t tim  = fCluster->GetStartTime();
	Double_t timo = ((CDCHit *)obj)->fCluster->GetStartTime();
   	if (tim < timo) return -1;
   	else if (tim > timo) return 1;
   	else return 0;
}

void CDCHit::DebugPrint(const Char_t *opt)
{
	if (!fDone) Update();
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< " pk = " << setw(5) << setprecision(1) << fCluster->GetPeak()
		<< " (X,Y) = (" << setw(5) << setprecision(3) << fWire->X()
		<< "," << setw(6) << setprecision(3) << fWire->Y() << ")"
		<< " d = "  << setw(6) << setprecision(3) << fXdrift
		<< " dd = " << setw(6) << setprecision(4) << fDXdrift
		<< endl;
	if (strstr(opt,"Detailed")) {
		cerr 	<< "    "
			<< " P_A = " << setw(6) << setprecision(1) 
				  	  << fCluster->GetSum() << endl;
    	}
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);
}

void CDCHit::Update()
{
	if (fDone) return;
	fDone = kTRUE;
	CalculatePosition();
}

void CDCHit::CalculatePosition()
{
	if (fWire) {
		fXdrift  = fWire->fCalib->X(fCluster);
		fDXdrift = fWire->fCalib->DX(fCluster);
	} else {
		cerr 	<< "::::: Error in CDCHit::CalculatePosition"  << endl
			<< " This hit belong to no CDCHitWire object!" << endl
			<< " Append this to some CDCHitWire object!"   << endl;
	}
}


//_____________________________________________________________________
// --------
// Hit Wire
// --------

ClassImp(CDCHitWire)

CDCHitWire::CDCHitWire(CDCTimeToXRelation *calib) : TObjArray(1),
	fLayerNo(0), fCellNo(0), fWireNo(0), 
	fXwire(0.), fYwire(0.), fCell(0),
	fCalib(calib) {}

CDCHitWire::~CDCHitWire()
{
	Delete();
}

Bool_t CDCHitWire::IsSortable() const { return kTRUE; }

Int_t CDCHitWire::Compare(TObject *obj)
{
	Int_t wire  = fWireNo;
	Int_t wireo = ((CDCHitWire *)obj)->fWireNo;
   	if (wire < wireo) return -1;
   	else if (wire > wireo) return 1;
   	else return 0;
}

Int_t CDCHitWire::Append(CDCHit *ht)
{
	Int_t 	 layer, cell, wire;
	Double_t xwire, ywire;
	ChannelToWire(ht->GetCluster(),layer,cell,wire,xwire,ywire);
	
	// Return value = -1:	1st hit.
	//		=  0:	successfully appended.
	//		=  1:	new wire.
	//		=  2:	new cell.
	//		=  3:	new layer.
	
	if (GetEntries() == 0) {
		fLayerNo  = layer;
		fCellNo	  = cell;
		fWireNo	  = wire;
		fXwire	  = xwire;
		fYwire	  = ywire;
		ht->fWire = this;
                Add(ht);
                return -1;
	} else {
		if (layer == fLayerNo) {
			if (cell == fCellNo) {
                    		if (wire == fWireNo) {
                    			ht->fWire = this;
                        		Add(ht);
                        		return 0;
                        	} else {
                        		return 1;
                        	}
                        } else {
                        	return 2;
                        }
                } else {
                        return 3;
                }
        }
}

void CDCHitWire::DebugPrint(const Char_t *opt)
{
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< "wire = (" << setw(1) << fLayerNo
		<< "," << setw(1) << fCellNo
		<< "," << setw(1) << fWireNo << ")";
	TIter next(this);
	CDCHit *hp;
	Int_t n = 0;
	while ((hp = (CDCHit *)next())) {
		if (n != 0) cerr << "                  ";
		cerr << setw(3) << ++n << " ";
		hp->DebugPrint(opt);
	}
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);
}


void CDCHitWire::ChannelToWire(CDCCluster *cluster,
  			       Int_t &layer, Int_t &cell, Int_t &wire, 
  			       Double_t &x, Double_t &y)
{
	Int_t cr  = 0;		// We have only 1 crate now.
	Int_t st  = cluster->GetStationNo();
	Int_t ch  = cluster->GetChannelNo();
	Int_t dch, id;
	
	CDCRawData     *rd = (CDCRawData *)gJSF->FindModule("CDCRawData");
	CDCBeginRunBuf *br = rd->GetBeginRunBuf();
	Int_t fmtid = br->GetHeader()->GetFormatID();
	switch (fmtid) {
	    case 0: {					// 98 format 
	        dch = ch;
		break;
	    }
	    case 1: {					// new format
	        br->EChannelToDChannel(cr,st,ch,id,dch);
		break;
	    }
	    default: {
	        cerr	<< "////// ERROR in CDCHitWire::ChannelToWire" << endl 
	        	<< "Invalid format ID = " << fmtid << endl;
		break;
	    }
	}
	ChannelToWire(dch,layer,cell,wire,x,y);	
}


void CDCHitWire::ChannelToWire(Int_t dch,
  			       Int_t &layer, Int_t &cell, Int_t &wire, 
  			       Double_t &x, Double_t &y)
{
//////////////// Very Temporary ////////////////////
/////	Should be moved to data fields  ////////////
/////   The values are only approximate ////////////

	Int_t nwirepercell   = 5;
	Double_t cellheight  = 10.;
	Double_t cellwidth   = 5.;
	Double_t wirespacing = 1.;

/////// This works only for baby chamber ///////////

	Int_t c  = dch/nwirepercell;
//	layer    = c%2;
	layer    = (c+1)/2;
	cell     = c/2;
	wire     = dch%nwirepercell;
//	x        = (1-c)*cellwidth;  
	x        = (c > 1 ? -1 : c)*cellwidth;  
        y        = layer*cellheight + wire*wirespacing;

//////////////// Very Temporary ////////////////////
}
//_____________________________________________________________________
// --------
// Hit Cell
// --------

ClassImp(CDCHitCell)

CDCHitCell::CDCHitCell() : TObjArray(1),
	fLayerNo(0), fCellNo(0),
	fLayer(0) {}

CDCHitCell::~CDCHitCell()
{
	Delete();
}

Bool_t CDCHitCell::IsSortable() const { return kTRUE; }

Int_t CDCHitCell::Compare(TObject *obj)
{
	Int_t cell  = fCellNo;
	Int_t cello = ((CDCHitCell *)obj)->fCellNo;
   	if (cell < cello) return -1;
   	else if (cell > cello) return 1;
   	else return 0;
}

Int_t CDCHitCell::Append(CDCHitWire *wire)
{	
	Int_t layer = wire->GetLayerNo();
	Int_t cell  = wire->GetCellNo();
	
	// Return value = -1:	new wire.
	//		=  0:	successfully appended.
	//		=  1:	new cell.
	//		=  2:	new layer.
	
	if (GetEntries() == 0) {
		fLayerNo    = layer;
		fCellNo	    = cell;
		wire->fCell = this;
                Add(wire);
                return 0;
	} else {
		if (layer == fLayerNo) {
			if (cell == fCellNo) {
				wire->fCell = this;
                        	Add(wire);
                        	return 0;
                        } else {
                        	return 1;
                        }
                } else {
                        return 2;
                }
        }
}

void CDCHitCell::DebugPrint(const Char_t *opt)
{
	cerr 	<< "(layer,cell) = " 
		<< "(" << fLayerNo << "," << fCellNo << ")" << endl;
	TIter next(this);
	CDCHitWire *wire;
	Int_t n = 0;
	while ((wire = (CDCHitWire *)next())) {
		cerr << setw(3) << ++n << " ";
		wire->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// ---------
// Hit Layer
// ---------

ClassImp(CDCHitLayer)

CDCHitLayer::CDCHitLayer() : TObjArray(1), fLayerNo(0) {}

CDCHitLayer::~CDCHitLayer()
{
	Delete();
}

Bool_t CDCHitLayer::IsSortable() const { return kTRUE; }

Int_t CDCHitLayer::Compare(TObject *obj)
{
	Int_t layer  = fLayerNo;
	Int_t layero = ((CDCHitLayer *)obj)->fLayerNo;
   	if (layer < layero) return -1;
   	else if (layer > layero) return 1;
   	else return 0;
}

Int_t CDCHitLayer::Append(CDCHitCell *cell)
{	
	Int_t layer = cell->GetLayerNo();
	
	// Return value = -1:	new cell.
	//		=  0:	successfully appended.
	//		=  1:	new layer.
	
	if (GetEntries() == 0) {
		fLayerNo     = layer;
		cell->fLayer = this;
                Add(cell);
                return 0;
	} else {
		if (layer == fLayerNo) {
			cell->fLayer = this;
                        Add(cell);
                        return 0;
                } else {
                        return 1;
                }
        }
}

void CDCHitLayer::DebugPrint(const Char_t *opt)
{
	cerr 	<< "Layer = " << fLayerNo << endl;
	TIter next(this);
	CDCHitCell *cell;
	while ((cell = (CDCHitCell *)next())) {
		cell->DebugPrint(opt);
	}
}



//_____________________________________________________________________
// ----------
// Hit Buffer
// ----------

ClassImp(CDCHitBuf)

CDCHitBuf::CDCHitBuf(CDCTimeToXRelation *calib,
			   const char *name,const char *title, 
			   CDCHitMaker *module) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), xWidth(100.), xPeak(10.), xSum(1000.),
	fClusterData(0), fHitLayers(1), fCalib(calib) {}

CDCHitBuf::CDCHitBuf(CDCHitMaker *module,
			   CDCTimeToXRelation *calib,
			   const char *name, const char *title) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), xWidth(100.), xPeak(10.), xSum(1000.),
	fClusterData(0), fHitLayers(1), fCalib(calib) {}

CDCHitBuf::~CDCHitBuf()
{
	fHitLayers.Delete();
}

void CDCHitBuf::SetClusterData(CDCClusterBuf *clbuf)
{
	fDone        = kFALSE;
	fClusterData = clbuf;
	if (fHitLayers.GetEntries() > 0) fHitLayers.Delete();
}

void CDCHitBuf::SetCalibrationData(CDCTimeToXRelation *calib)
{
	fDone  = kFALSE;
	fCalib = calib;
}

Bool_t CDCHitBuf::IsInitialized()  const
{
	if (fClusterData) return kTRUE;
	else return kFALSE;
}

Int_t CDCHitBuf::GetNlayers()
{
	if (!fDone) MakeHits(); 
	return fHitLayers.GetEntries();
}

TObjArray * CDCHitBuf::GetHitLayers()
{
	if (!fDone) MakeHits();
	return &fHitLayers;
}

void CDCHitBuf::MakeHits()
{
	if (!IsInitialized()) {
		cerr << "ERROR  CDCHitBuf::MakeHits" << endl
		     << "...... cluster data not yet initialized." << endl;
		return;
	} else if (fDone) {
		return;
	}
	
	fDone = kTRUE;
	
	CDCHitLayer *layerp = 0;
	CDCHitCell  *cellp  = 0;
	CDCHitWire  *wirep  = 0;
	
	TIter next(fClusterData->GetClusters());
	CDCCluster *clp;
	while ((clp = (CDCCluster *)next())) {
		Double_t w   = clp->GetWidth();
		Double_t pk  = clp->GetPeak();
		Double_t sum = clp->GetSum();
		if (w > xWidth && pk > xPeak && sum > xSum) {
			if (!layerp) {
				layerp = new CDCHitLayer();
				cellp  = new CDCHitCell();
				wirep  = new CDCHitWire(fCalib);
			}
			CDCHit *ht = new CDCHit(clp);
			Int_t r1 = wirep->Append(ht);
			switch (r1) {
				case -1: {		// 1st hit
					cellp->Append(wirep);
					layerp->Append(cellp);
					fHitLayers.Add(layerp);
					break;
				}
				case  0: break; 	// appended
				case  3: {		// new layer
					wirep = new CDCHitWire(fCalib);
					wirep->Append(ht);
					cellp = new CDCHitCell();
					cellp->Append(wirep);
					layerp = new CDCHitLayer();
					layerp->Append(cellp);
					fHitLayers.Add(layerp);
					break;
				}
				case  2: {		// new cell
					wirep = new CDCHitWire(fCalib);
					wirep->Append(ht);
					cellp = new CDCHitCell();
					cellp->Append(wirep);
					layerp->Append(cellp);
					break;
				}
				case  1: {		// new wire
					wirep = new CDCHitWire(fCalib);
					wirep->Append(ht);
					cellp->Append(wirep);
					break;
				}
			}
		}
	}
}

void CDCHitBuf::DebugPrint(const Char_t *opt)
{
	cerr	<< fHitLayers.GetEntries()
		<< " layers in this event:" << endl;
	TIter next(&fHitLayers);
	CDCHitLayer *lp;
	while ((lp = (CDCHitLayer *)next())) {
		lp->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// ---------
// Hit Maker
// ---------

ClassImp(CDCHitMaker)

CDCHitMaker::CDCHitMaker(CDCTimeToXRelation *calib,
			       const char *name, const char *title)
       : JSFModule(name,title), 
         fToDeleteCalib(calib ? kFALSE : kTRUE), fCalib(calib) {}

CDCHitMaker::~CDCHitMaker()
{
	if (fToDeleteCalib && fCalib) delete fCalib;
}

void CDCHitMaker::SetCalibrationData(CDCTimeToXRelation *calib)
{
	fCalib = calib;
	fToDeleteCalib = kFALSE;
	((CDCHitBuf *)EventBuf())->SetCalibrationData(calib);
}

Bool_t CDCHitMaker::Initialize()
{
	if (!fCalib) {
		fCalib = new CDCTimeToXRelation();
		fToDeleteCalib = kTRUE;
	}
	fEventBuf = new CDCHitBuf(this,fCalib);
	
	return kTRUE;
}

Bool_t CDCHitMaker::BeginRun(Int_t run)
{
	fCalib->Update();
	
	return kTRUE;
}

Bool_t CDCHitMaker::Process(Int_t ev)
{
	CDCClusterMaker *clmp
		= (CDCClusterMaker*)gJSF->FindModule("CDCClusterMaker");
	CDCClusterBuf *clbufp = (CDCClusterBuf *)clmp->EventBuf();

	CDCHitBuf  *phbufp = (CDCHitBuf *)EventBuf();
	phbufp->SetClusterData(clbufp);	
	phbufp->MakeHits();

	return kTRUE;
}





