//*************************************************************************
//* -------------------
//* CDCTrackMaker Class
//* -------------------
//* (Description)
//*	This set of classes makes tracks.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCHit:
//*	CDCHitBuf:
//*	CDCHitMaker:
//* (Provides)
//*	CDCTrackMaker		: steering class
//*	CDCTrackBuf		: track data buffer class
//*	CDCTrack		: track data class
//*	CDCTrackHit		: track hit
//* (Update Record)
//*	2000/01/07  K.Fujii	Original version.
//*     2000/09/30  K.Fujii	Fixed a bug: nextinlst hasn't been reset!
//*				Quality check before poisoning hits.
//*	2000/10/02  K.Fujii	Loop over pivot combinations.
//*	2000/10/04  K.Fujii	Cell boundary check.
//*************************************************************************

#include "CDCTrackMaker.h"

//_____________________________________________________________________
// ---------
// Track Hit
// ---------
ClassImp(CDCTrackHit)

CDCTrackHit::CDCTrackHit() {}

CDCTrackHit::CDCTrackHit(CDCHit *ht, Int_t lr) :
	fResidual(0.), fLeftRight(lr), fHit(ht) {}

CDCTrackHit::CDCTrackHit(CDCTrackHit &hit) : Lockable(hit),
	fResidual(hit.fResidual), fLeftRight(hit.fLeftRight), fHit(hit.fHit) {}

CDCTrackHit::~CDCTrackHit() {}


//_____________________________________________________________________
// -----------
// Track Datum
// -----------

ClassImp(CDCTrack)

CDCTrack::CDCTrack() :
	fDone(kFALSE), fTrackType(0), fChi2(0.),
	fa(0.), fb(0.), fEaa(0.), fEba(0.), fEbb(0.) {}

CDCTrack::CDCTrack(CDCTrack &t) : TObjArray(0), Lockable(t),
	fDone(t.fDone), fTrackType(t.fTrackType), fChi2(t.fChi2),
	fa(t.fa), fb(t.fb), fEaa(t.fEaa), fEba(t.fEba), fEbb(t.fEbb)
{
	TIter next(&t);
	CDCTrackHit *hitp;
	while ((hitp = (CDCTrackHit *)next())) {
		Add(new CDCTrackHit(*hitp));
	}
}

CDCTrack::~CDCTrack() { if (GetEntries()) Delete(); }

Bool_t CDCTrack::IsSortable() const { return kTRUE; }

Int_t CDCTrack::Compare(TObject *obj)
{
	if (!fDone) Update();
	Double_t chi2 = ((CDCTrack *)obj)->fChi2;
   	if (fChi2 < chi2) return -1;
   	else if (fChi2 > chi2) return 1;
   	else return 0;
}

void CDCTrack::Add(TObject *obj)
{
	TObjArray::Add(obj);
	CDCTrackHit *hitp = (CDCTrackHit *)obj;
	if (!hitp->IsLocked()) fDone = kFALSE;
}

void CDCTrack::SetTrackType(Int_t type)
{
	fTrackType = type;
}

Int_t CDCTrack::GetTrackType()
{
	if (!fDone) Update();
	return fTrackType;
}

Double_t CDCTrack::GetChi2()
{
	if (!fDone) Update();
	return fChi2;
}

Double_t CDCTrack::Xat(Double_t y)
{
	if (!fDone) Update();
	return (fa*y + fb);
}

Double_t CDCTrack::SGX2at(Double_t y)
{
	if (!fDone) Update();
	return (fEaa*y*y + 2*fEba*y + fEbb);
}

void CDCTrack::DebugPrint(const Char_t *opt)
{
	if (!fDone) Update();
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< " Type = " << setw(2) << fTrackType
		<< " chi2 = " << setw(6) << setprecision(3) << fChi2
		<< " (a,b) = (" << setw(6) << setprecision(3) << fa
		<< "," << setw(6) << setprecision(3) << fb << ")"
		<< endl;
	if (strstr(opt,"Detailed")) {
		TIter next(this);
		CDCTrackHit *hitp;
		Int_t i = 0;
		while((hitp = (CDCTrackHit *)next())) {
			cerr << setw(2) << i++
			     << " L/R = " << setw(2) << hitp->fLeftRight
			     << " dx = "
			     << setw(6) << setprecision(3) << hitp->fResidual;
			hitp->fHit->DebugPrint();
		}
    	}
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);
}

void CDCTrack::Update()
{
	fDone = kTRUE;
	Fit();
}

void CDCTrack::Fit()
{
	Double_t det, x, y, rs2;
	Double_t syy = 0.;
	Double_t sy  = 0.;
	Double_t s   = 0.;
	Double_t sxy = 0.;
	Double_t sx  = 0.;
	TIter next(this);
	CDCTrackHit *hitp;
	while ((hitp = (CDCTrackHit *)next())) {
		CDCTrackHit &hit = *hitp;
		CDCHit      &ht  = *hit.fHit;
		if (hit.IsLocked()) continue;
		rs2  = 1./TMath::Power(ht.DXdrift(),2);
		x    = ht.Xwire() + (2*hit.fLeftRight - 1) * ht.Xdrift();
		y    = ht.Ywire();
		syy += rs2*y*y;
		sy  += rs2*y;
		s   += rs2;
		sxy += rs2*x*y;
		sx  += rs2*x;
	}
	det  = syy*s -sy*sy;
	fa   = (  s*sxy -  sy*sx)/det;
	fb   = (-sy*sxy + syy*sx)/det;
	fEaa =   s/det;
	fEba = -sy/det;
	fEbb = syy/det;
	
	fChi2 = 0.;
	next.Reset();
	while ((hitp = (CDCTrackHit *)next())) {
		CDCTrackHit &hit = *hitp;
		CDCHit      &ht  = *hit.fHit;
		x    = ht.Xwire() + (2*hit.fLeftRight - 1) * ht.Xdrift();
		y    = ht.Ywire();
	        hit.fResidual = x - fa*y - fb;
	        fChi2 += TMath::Power(hit.fResidual/ht.DXdrift(),2);
	}
}

//_____________________________________________________________________
// ------------
// Track Buffer
// ------------

ClassImp(CDCTrackBuf)

CDCTrackBuf::CDCTrackBuf(const char *name, const char *title, 
		         CDCTrackMaker *module) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fHitData(0), fTracks(1),
	xNsigmas(4.), xChi2(45.), xDXpivot(0.05), xKCNsigmas(2.) {}

CDCTrackBuf::CDCTrackBuf(CDCTrackMaker *module,
		     const char *name, const char *title) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fHitData(0), fTracks(1),
	xNsigmas(4.), xChi2(45.), xDXpivot(0.05), xKCNsigmas(2.) {}

CDCTrackBuf::~CDCTrackBuf()
{
	if (fTracks.GetEntries() > 0) fTracks.Delete();
}

void CDCTrackBuf::SetHitData(CDCHitBuf *htbuf)
{
	fDone    = kFALSE;
	fHitData = htbuf;
	if (fTracks.GetEntries() > 0) fTracks.Delete();
}

Bool_t CDCTrackBuf::IsInitialized()  const
{
	if (fHitData) return kTRUE;
	else return kFALSE;
}

Int_t CDCTrackBuf::GetNtracks()
{
	if (!fDone) MakeTracks(); 
	return fTracks.GetEntries();
}

TObjArray * CDCTrackBuf::GetTracks()
{
	if (!fDone) MakeTracks();
	return &fTracks;
}

void CDCTrackBuf::MakeTracks()
{
	if (!IsInitialized()) {
		cerr << "ERROR  CDCTrackBuf::MakeTracks" << endl
		     << "...... hit data not yet initialized." << endl;
		return;
	} else if (fDone) {
		return;
	}
	
	fDone = kTRUE;

	// Form vector hits (track segments in a single cell) 
	// (Caution)
	// 	Vector hits are currently stored as ordinary tracks
	//	in the track buffer fTracks.
	
	TIter nextLayer(fHitData->GetHitLayers());
	CDCHitLayer *lyrp;
	while ((lyrp = (CDCHitLayer *)nextLayer())) {
		TIter nextCell(lyrp);
		CDCHitCell *celp;
		while ((celp = (CDCHitCell *)nextCell())) {
			MakeVectorHits(celp);
		}
	}
	// One should combine vector hits to form global tracks here.
}

void CDCTrackBuf::MakeVectorHits(CDCHitCell *celp)
{
     ////// Very Temporary Treatment //////
     Int_t    lrsel     = (celp->GetLayerNo() == 1 ? 0 : 1);
     Double_t cellwidth  = 5.0;
     Double_t cellmargin = 0.1;
     ////// Very Temporary Treatment //////
	
     Int_t nwires = celp->GetEntries();
     if (nwires < 4)	return;

     celp->Sort();	// sort wires
	
     Int_t pivotcomb;
     for (pivotcomb = 0; pivotcomb < (nwires < 5 ? 1 : 3); pivotcomb++) {
	Int_t iw1st;
	Int_t iwlst;
	switch (pivotcomb) {
	   case 1:
	      iw1st = 0;
	      iwlst = nwires - 2;
	      break;
	   case 2:
	      iw1st = 1;
	      iwlst = nwires - 1;
	      break;
	   case 0:
	   default:
	      iw1st = 0;
	      iwlst = nwires - 1;
	      break;
	}	      
	CDCHitWire *wire1st = (CDCHitWire *)celp->At(iw1st);
	CDCHitWire *wirelst = (CDCHitWire *)celp->At(iwlst);
	
	TIter nextin1st(wire1st);
	TIter nextinlst(wirelst);
	
	CDCHit *ht1st;
	CDCHit *htmdl;
	CDCHit *htlst;
		
	while ((ht1st = (CDCHit *)nextin1st())) {	// hits in 1st wire
	   if (ht1st->IsLocked()) continue;	 	// skip if locked
	   Double_t x1  = ht1st->Xdrift();
	   if (x1 < -cellmargin || x1 > cellwidth+cellmargin) continue;
	   Double_t dx1 = ht1st->DXdrift();
	   if (dx1 > xDXpivot) continue;		// skip if bad quality
	   Int_t lr1st  = lrsel;			// TEMPORARY
	   nextinlst.Reset();				// reset iterator
	   while ((htlst = (CDCHit *)nextinlst())) {	// hits in last wire
	      if (htlst->IsLocked()) continue;		// skip if locked
	      Double_t x2  = htlst->Xdrift();
	      if (x2 < -cellmargin || x2 > cellwidth+cellmargin) continue;
	      Double_t dx2 = htlst->DXdrift();
	      if (dx2 > xDXpivot) continue;		// skip if bad quality
	      Int_t lrlst  = lrsel;			// TEMPORARY

	      CDCTrack *trialtkp = new CDCTrack();	   // make trial track
	      trialtkp->Add(new CDCTrackHit(ht1st,lr1st)); // add 1st hit
	      trialtkp->Add(new CDCTrackHit(htlst,lrlst)); // add lst hit

	      Int_t i;
	      for (i = 0; i < nwires; i++) { // loop over middle wires
	         if (i == iw1st || i == iwlst) continue;  // skip pivots
	         Double_t diffmn = 999999.;
	         CDCHitWire *wiremdl = (CDCHitWire *)celp->At(i);
	         TIter nextinmdl(wiremdl);
	         CDCHit *bst = 0;
	         Int_t lrmdl = lrsel;			  // TEMPORARY
	         while ((htmdl = (CDCHit *)nextinmdl())) {// hits in mdl wire
	            if (htmdl->IsLocked()) continue;	  // skip if locked
	            Double_t x3  = htmdl->Xdrift();
	            if (x3 < -cellmargin || x3 > cellwidth+cellmargin) continue;
	            Double_t dx3      = htmdl->DXdrift();
	            if (dx3 > xDXpivot) continue;	  // skip if bad quality
	            CDCTrackHit *hitp = new CDCTrackHit(htmdl,lrmdl);
	            hitp->Lock();		// exclude this hit from fit
		    trialtkp->Add(hitp);	// add this temporarily to track
		    trialtkp->Update();		// and update residual
	            Double_t diff   = hitp->GetResidual();
	            Double_t diffct = xNsigmas*(dx1+dx2+dx3);
		    trialtkp->Remove(hitp);	// remove this hit from track
		    delete hitp;		// delete this temp. hit
	            if (TMath::Abs(diff) > diffct) continue;	// resid. cut
	            if (TMath::Abs(diff) < diffmn) {	// select the best
	               diffmn = TMath::Abs(diff);
	               bst    = htmdl;
	            }
	         }
	         if (bst) {			// add the best to track
	            trialtkp->Add(new CDCTrackHit(bst,lrmdl));
	         }
	      }
	      if (trialtkp->GetEntries() < 4 || trialtkp->GetChi2() > xChi2) {
	         delete trialtkp;			// track rejected
	      } else {					// track accepted
	         fTracks.Add(trialtkp);			// added to track buffer
	         TIter next(trialtkp);
	         CDCTrackHit *ht;
	         Int_t ngoodhits = 0;
	         while ((ht = (CDCTrackHit *)next())) { // poison used hits
	            if (TMath::Power(ht->GetResidual()/xKCNsigmas,2) <
		    	trialtkp->SGX2at(ht->GetCDCHit()->Ywire())) {
	            	ht->GetCDCHit()->Lock();
	            	ngoodhits++;
		    }
	         }
	         if (ngoodhits >=4) goto NEXT;
	      }
	   }
	   NEXT: continue;
	}
     }
}

void CDCTrackBuf::DebugPrint(const Char_t *opt)
{
	cerr	<< fTracks.GetEntries()
		<< " tracks in this event:" << endl;
	TIter next(&fTracks);
	CDCTrack *tp;
	Int_t n = 0;
	while ((tp = (CDCTrack *)next())) {
		cerr << setw(3) << ++n << " ";
		tp->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// -----------
// Track Maker
// -----------

ClassImp(CDCTrackMaker)

CDCTrackMaker::CDCTrackMaker(const char *name, const char *title)
       : JSFModule(name,title) {}

CDCTrackMaker::~CDCTrackMaker() {}

Bool_t CDCTrackMaker::Initialize()
{
	fEventBuf = new CDCTrackBuf(this);	
	return kTRUE;
}

Bool_t CDCTrackMaker::BeginRun(Int_t run)
{
	return kTRUE;
}

Bool_t CDCTrackMaker::Process(Int_t ev)
{
	CDCHitMaker *htmp
		= (CDCHitMaker*)gJSF->FindModule("CDCHitMaker");
	CDCHitBuf *htbufp = (CDCHitBuf *)htmp->EventBuf();

	CDCTrackBuf  *tkbufp = (CDCTrackBuf *)EventBuf();
	tkbufp->SetHitData(htbufp);	
	tkbufp->MakeTracks();

	return kTRUE;
}




